@@ -4,14 +4,18 @@ import pathlib
 import sqlite3
 import tempfile
 import uuid
+from typing import List
 
 import msgpack
-from kodexa.model import Document, ContentNode, SourceMetadata
-from kodexa.model.model import ContentClassification, DocumentMetadata, ContentFeature
 
-logger = logging.getLogger('kodexa.model.persistence')
+from kodexa.model import Document, ContentNode, SourceMetadata
+from kodexa.model.model import ContentClassification, DocumentMetadata, ContentFeature, ContentException
+
+logger = logging.getLogger()
 
 # Heavily used SQL
+EXCEPTION_INSERT = "INSERT INTO content_exceptions (tag, message, exception_details, group_uuid, tag_uuid, exception_type, severity, node_uuid) VALUES (?, ?, ?, ?, ?, ?, ?, ?)"
+EXCEPTION_SELECT = "select tag, message, exception_details, group_uuid, tag_uuid, exception_type, severity, node_uuid from content_exceptions"
 
 FEATURE_INSERT = "INSERT INTO ft (id, cn_id, f_type, binary_value, single, tag_uuid) VALUES (?,?,?,?,?,?)"
 FEATURE_DELETE = "DELETE FROM ft where cn_id=? and f_type=?"
@@ -52,7 +56,8 @@ class SqliteDocumentPersistence(object):
                 # At this point we need to load the db
                 self.is_new = False
         else:
-            new_file, filename = tempfile.mkstemp(suffix='.kddb')
+            from kodexa import KodexaPlatform
+            new_file, filename = tempfile.mkstemp(suffix='.kddb', dir=KodexaPlatform.get_tempdir())
             self.is_tmp = True
 
         self.current_filename = filename
@@ -63,6 +68,107 @@ class SqliteDocumentPersistence(object):
         self.cursor.execute("pragma temp_store = memory")
         self.cursor.execute("pragma mmap_size = 30000000000")
 
+    def get_all_tags(self):
+        "select * from cn where id in (select cn_id from ft where f_type in (select id from f_type where name like 'tag:%'))"
+        features = []
+        for feature in self.cursor.execute(
+                "select name from f_type where name like 'tag:%'").fetchall():
+            features.append(feature[0].split(':')[1])
+
+        return features
+
+    def update_features(self, node):
+
+        next_feature_id = self.get_max_feature_id()
+        all_features = []
+        for feature in node.get_features():
+            binary_value = sqlite3.Binary(msgpack.packb(feature.value, use_bin_type=True))
+
+            tag_uuid = None
+            if feature.feature_type == 'tag' and 'uuid' in feature.value[0]:
+                tag_uuid = feature.value[0]['uuid']
+
+            all_features.append(
+                [next_feature_id, node.uuid, self.get_feature_type_id(feature),
+                 binary_value, feature.single, tag_uuid])
+
+            next_feature_id = next_feature_id + 1
+
+        self.cursor.execute("DELETE FROM ft where cn_id=?", [node.uuid])
+        self.cursor.executemany(FEATURE_INSERT, all_features)
+
+    def update_node(self, node):
+        self.cursor.execute('update cn set idx=?, pid=? where id=?',
+                            [node.index, node._parent_uuid,
+                             node.uuid])
+
+    def get_content_nodes(self, node_type, parent_node: ContentNode, include_children):
+        nodes = []
+
+        results = []
+        if include_children:
+
+            if node_type == "*":
+                query = """
+                            with recursive
+                            parent_node(id, pid, nt, idx, path) AS (
+                                VALUES (?,?,?,?,?)
+                                UNION ALL
+                                SELECT cns.id, cns.pid, cns.nt, cns.idx, parent_node.path || substr('0000000' || cns.idx, -6, 6) 
+                                FROM cn cns, parent_node
+                                WHERE parent_node.id = cns.pid  
+                            )
+                            SELECT id, pid, nt, idx, path from parent_node order by path
+                            """
+
+                try:
+                    results = self.cursor.execute(query,
+                                                  [parent_node.uuid,
+                                                   parent_node.get_parent().uuid if parent_node.get_parent() else None,
+                                                   next(key for key, value in self.node_types.items() if
+                                                        value == parent_node.get_node_type()),
+                                                   parent_node.index, f"{parent_node.index}".zfill(6)]).fetchall()
+                except StopIteration:
+                    return []
+            else:
+                query = """
+                                with recursive
+                                parent_node(id, pid, nt, idx, path) AS (
+                                    VALUES (?,?,?,?,?)
+                                    UNION ALL
+                                    SELECT cns.id, cns.pid, cns.nt, cns.idx, parent_node.path || substr('000000' || cns.idx, -6, 6) 
+                                    FROM cn cns, parent_node
+                                    WHERE parent_node.id = cns.pid  
+                                )
+                                SELECT id, pid, nt, idx, path from parent_node where nt=? order by path
+                                """
+
+                try:
+                    results = self.cursor.execute(query,
+                                                  [parent_node.uuid,
+                                                   parent_node.get_parent().uuid if parent_node.get_parent() else None,
+                                                   next(key for key, value in self.node_types.items() if
+                                                        value == parent_node.get_node_type()),
+                                                   parent_node.index,
+                                                   f"{parent_node.index}".zfill(6),
+                                                   next(key for key, value in self.node_types.items() if
+                                                        value == node_type)]).fetchall()
+                except StopIteration:
+                    return []
+        else:
+            query = "select id, pid, nt, idx from cn where pid=? and nt=? order by idx"
+            try:
+                results = self.cursor.execute(query,
+                                              [parent_node.uuid, next(key for key, value in self.node_types.items() if
+                                                                      value == node_type)]).fetchall()
+            except StopIteration:
+                return []
+
+        for raw_node in list(results):
+            nodes.append(self.__build_node(raw_node))
+
+        return nodes
+
     def initialize(self):
         if self.is_new:
             self.__build_db()
@@ -71,16 +177,17 @@ class SqliteDocumentPersistence(object):
 
     def close(self):
         if self.is_tmp or self.delete_on_close:
+            pathlib.Path(self.current_filename).unlink()
+        else:
             self.cursor.close()
             self.connection.close()
-            pathlib.Path(self.current_filename).unlink()
 
     def get_max_feature_id(self):
         max_id = self.cursor.execute("select max(id) from ft").fetchone()
         if max_id[0] is None:
             return 1
-        else:
-            return max_id[0] + 1
+
+        return max_id[0] + 1
 
     def __build_db(self):
         self.cursor.execute("CREATE TABLE metadata (id integer primary key, metadata text)")
@@ -107,6 +214,19 @@ class SqliteDocumentPersistence(object):
         self.cursor.execute("CREATE INDEX cnp_perf ON cnp(cn_id, pos);")
         self.cursor.execute("CREATE INDEX f_perf ON ft(cn_id);")
         self.cursor.execute("CREATE INDEX f_perf2 ON ft(tag_uuid);")
+        self.cursor.execute("""CREATE TABLE content_exceptions
+                                    (
+                                        id           integer primary key,
+                                        tag          text,
+                                        message      text,
+                                        exception_details text,
+                                        group_uuid   text,
+                                        tag_uuid     text,
+                                        exception_type text,
+                                        severity     text,
+                                        node_uuid    text
+                                    )""")
+        self.document.version = "4.0.2"
 
         self.__update_metadata()
 
@@ -173,8 +293,8 @@ class SqliteDocumentPersistence(object):
                 self.cursor.executemany(CONTENT_NODE_PART_INSERT, cn_parts_values)
 
             return ([cn_values], cn_parts_values)
-        else:
-            raise "Node must have a UUID?"
+
+        raise Exception("Node must have a UUID?")
 
     def __clean_none_values(self, d):
         clean = {}
@@ -207,9 +327,6 @@ class SqliteDocumentPersistence(object):
 
         metadata = msgpack.unpackb(self.cursor.execute("select * from metadata").fetchone()[1])
         self.document.metadata = DocumentMetadata(metadata['metadata'])
-        for mixin in metadata['mixins']:
-            from kodexa.mixins import registry
-            registry.add_mixin_to_document(mixin, self.document)
         self.document.version = metadata['version'] if 'version' in metadata and metadata[
             'version'] else Document.PREVIOUS_VERSION  # some older docs don't have a version or it's None
 
@@ -219,6 +336,8 @@ class SqliteDocumentPersistence(object):
             self.document.source = SourceMetadata.from_dict(metadata['source'])
         if 'labels' in metadata and metadata['labels']:
             self.document.labels = metadata['labels']
+        if 'mixins' in metadata and metadata['mixins']:
+            self.document._mixins = metadata['mixins']
         if 'taxomomies' in metadata and metadata['taxomomies']:
             self.document.taxonomies = metadata['taxomomies']
         if 'classes' in metadata and metadata['classes']:
@@ -231,7 +350,7 @@ class SqliteDocumentPersistence(object):
             self.document.content_node = self.__build_node(
                 root_node)
 
-        if self.document.version != '4.0.1':
+        if self.document.version != '4.0.1' and self.document.version != '4.0.2':
             # We need to migrate this to a 4.0.1 document
             self.cursor.execute("""CREATE TABLE ft
                                     (
@@ -242,7 +361,8 @@ class SqliteDocumentPersistence(object):
                                         single       integer,
                                         tag_uuid     text
                                     )""")
-            self.cursor.execute("insert into ft select f.id, f.cn_id, f.f_type, fv.binary_value, fv.single, null from f, f_value fv where fv.id = f.fvalue_id")
+            self.cursor.execute(
+                "insert into ft select f.id, f.cn_id, f.f_type, fv.binary_value, fv.single, null from f, f_value fv where fv.id = f.fvalue_id")
             # we will create a new feature table
             self.cursor.execute("drop table f")
             self.cursor.execute("drop table f_value")
@@ -251,6 +371,21 @@ class SqliteDocumentPersistence(object):
             self.document.version = "4.0.1"
             self.update_metadata()
 
+        self.cursor.execute("""CREATE TABLE IF NOT EXISTS content_exceptions
+                                    (
+                                        id           integer primary key,
+                                        tag          text,
+                                        message      text,
+                                        exception_details text,
+                                        group_uuid   text,
+                                        tag_uuid     text,
+                                        exception_type text,
+                                        severity     text,
+                                        node_uuid    text
+                                    )""")
+        self.document.version = "4.0.2"
+        self.update_metadata()
+
     def get_content_parts(self, new_node):
         content_parts = self.cursor.execute(
             "select cn_id, pos, content, content_idx from cnp where cn_id = ? order by pos",
@@ -301,16 +436,16 @@ class SqliteDocumentPersistence(object):
         node_row = self.cursor.execute("select id, pid, nt, idx from cn where id = ?", [node_id]).fetchone()
         if node_row:
             return self.__build_node(node_row)
-        else:
-            return None
+
+        return None
 
     def get_parent(self, content_node):
 
         parent = self.cursor.execute("select pid from cn where id = ?", [content_node.uuid]).fetchone()
         if parent:
-            return self.__get_node(parent[0])
-        else:
-            return None
+            return self.get_node(parent[0])
+
+        return None
 
     def update_metadata(self):
         self.__update_metadata()
@@ -324,8 +459,7 @@ class SqliteDocumentPersistence(object):
         if self.document.content_node:
             self.__insert_node(self.document.content_node, None)
 
-    def get_bytes(self):
-
+    def sync(self):
         self.__update_metadata()
         self.cursor.execute("pragma optimize")
         self.connection.commit()
@@ -338,6 +472,8 @@ class SqliteDocumentPersistence(object):
         self.cursor.execute("pragma temp_store = memory")
         self.cursor.execute("pragma mmap_size = 30000000000")
 
+    def get_bytes(self):
+        self.sync()
         with open(self.current_filename, 'rb') as f:
             return f.read()
 
@@ -390,8 +526,40 @@ class SqliteDocumentPersistence(object):
         next_id = self.cursor.execute("select max(id) from cn").fetchone()
         if next_id[0] is None:
             return 1
+
+        return next_id[0] + 1
+
+    def get_tagged_nodes(self, tag, tag_uuid=None):
+        content_nodes = []
+        if tag_uuid is None:
+            query = f"select cn_id from ft where f_type in (select id from f_type where name like 'tag:{tag}')"
         else:
-            return next_id[0] + 1
+            query = f"select cn_id from ft where f_type in (select id from f_type where name like 'tag:{tag}') and tag_uuid = '{tag_uuid}'"
+        for content_node_ids in self.cursor.execute(
+                query).fetchall():
+            content_nodes.append(self.get_node(content_node_ids[0]))
+
+        return content_nodes
+
+    def add_exception(self, exception: ContentException):
+        # Add an exception to the exception table
+        self.cursor.execute(EXCEPTION_INSERT,
+                            [exception.tag, exception.message, exception.exception_details, exception.group_uuid,
+                             exception.tag_uuid, exception.exception_type, exception.severity, exception.node_uuid])
+
+    def get_exceptions(self) -> List[ContentException]:
+        exceptions = []
+        for exception in self.cursor.execute(EXCEPTION_SELECT).fetchall():
+            exceptions.append(ContentException(tag=exception[0], message=exception[1], exception_details=exception[2],
+                                               group_uuid=exception[3], tag_uuid=exception[4],
+                                               exception_type=exception[5],
+                                               severity=exception[6], node_uuid=exception[7]))
+        return exceptions
+
+    def replace_exceptions(self, exceptions: List[ContentException]):
+        self.cursor.execute("delete from content_exceptions")
+        for exception in exceptions:
+            self.add_exception(exception)
 
 
 class SimpleObjectCache(object):
@@ -406,11 +574,11 @@ class SimpleObjectCache(object):
         self.next_id = 1
         self.dirty_objs = set()
 
-    def get_obj(self, id):
-        if id in self.objs:
-            return self.objs[id]
-        else:
-            return None
+    def get_obj(self, obj_id):
+        if obj_id in self.objs:
+            return self.objs[obj_id]
+
+        return None
 
     def add_obj(self, obj):
         if obj.uuid is None:
@@ -422,12 +590,13 @@ class SimpleObjectCache(object):
     def remove_obj(self, obj):
         if obj.uuid in self.objs:
             self.objs.pop(obj.uuid)
-            self.dirty_objs.remove(obj.uuid)
+            if obj.uuid in self.dirty_objs:
+                self.dirty_objs.remove(obj.uuid)
 
     def get_dirty_objs(self):
         results = []
-        for id in set(self.dirty_objs):
-            node = self.get_obj(id)
+        for set_id in set(self.dirty_objs):
+            node = self.get_obj(set_id)
             if node is not None:
                 results.append(node)
         return results
@@ -455,26 +624,47 @@ class PersistenceManager(object):
 
         self._underlying_persistence = SqliteDocumentPersistence(document, filename, delete_on_close)
 
+    def add_exception(self, exception: ContentException):
+        self._underlying_persistence.add_exception(exception)
+
+    def get_exceptions(self) -> List[ContentException]:
+        return self._underlying_persistence.get_exceptions()
+
+    def replace_exceptions(self, exceptions: List[ContentException]):
+        self._underlying_persistence.replace_exceptions(exceptions)
+
+    def get_all_tags(self):
+        return self._underlying_persistence.get_all_tags()
+
+    def get_tagged_nodes(self, tag, tag_uuid=None):
+        self.flush_cache()
+        return self._underlying_persistence.get_tagged_nodes(tag, tag_uuid)
+
     def initialize(self):
         self._underlying_persistence.initialize()
 
         self.node_cache.next_id = self._underlying_persistence.get_next_node_id()
 
+    def get_parent(self, node):
+        if node.uuid in self.node_parent_cache:
+            return self.node_cache.get_obj(self.node_parent_cache[node.uuid])
+
+        return self._underlying_persistence.get_parent(node)
+
     def close(self):
         self._underlying_persistence.close()
 
-    def get_bytes(self):
-
+    def flush_cache(self):
         all_node_ids = []
         all_nodes = []
         all_content_parts = []
         all_features = []
         node_id_with_features = []
 
-        logger.info("Merging cache to persistance")
+        logger.debug("Merging cache to persistence")
         dirty_nodes = self.node_cache.get_dirty_objs()
 
-        logger.info(f"Identified {len(dirty_nodes)} nodes to update")
+        logger.debug(f"Identified {len(dirty_nodes)} nodes to update")
 
         next_feature_id = self._underlying_persistence.get_max_feature_id()
         for node in dirty_nodes:
@@ -502,17 +692,24 @@ class PersistenceManager(object):
 
                 self.node_cache.undirty(node)
 
-        logger.info(f"Writing {len(all_node_ids)} nodes")
+        logger.debug(f"Writing {len(all_node_ids)} nodes")
         self._underlying_persistence.cursor.executemany("DELETE FROM cn where id=?", all_node_ids)
         self._underlying_persistence.cursor.executemany("DELETE FROM ft where cn_id=?", node_id_with_features)
         self._underlying_persistence.cursor.executemany("INSERT INTO cn (pid, nt, idx, id) VALUES (?,?,?,?)", all_nodes)
         self._underlying_persistence.cursor.executemany("DELETE FROM cnp where cn_id=?", all_node_ids)
-        logger.info(f"Writing {len(all_content_parts)} content parts")
+        logger.debug(f"Writing {len(all_content_parts)} content parts")
 
         self._underlying_persistence.cursor.executemany(CONTENT_NODE_PART_INSERT, all_content_parts)
 
-        logger.info(f"Writing {len(all_features)} features")
+        logger.debug(f"Writing {len(all_features)} features")
         self._underlying_persistence.cursor.executemany(FEATURE_INSERT, all_features)
+
+    def get_content_nodes(self, node_type, parent_node, include_children):
+        return self._underlying_persistence.get_content_nodes(node_type, parent_node, include_children)
+
+    def get_bytes(self):
+        self.flush_cache()
+        self._underlying_persistence.sync()
         return self._underlying_persistence.get_bytes()
 
     def update_metadata(self):
@@ -547,7 +744,7 @@ class PersistenceManager(object):
 
             if node._parent_uuid not in self.child_cache:
                 self.child_cache[node._parent_uuid] = [node]
-                self.child_id_cache[node._parent_uuid] = set([node.uuid])
+                self.child_id_cache[node._parent_uuid] = {node.uuid}
             else:
                 if node.uuid not in self.child_id_cache[node._parent_uuid]:
                     self.child_id_cache[node._parent_uuid].add(node.uuid)
@@ -578,8 +775,21 @@ class PersistenceManager(object):
         self.node_cache.remove_obj(node)
 
         if node.uuid in self.node_parent_cache:
-            self.child_cache[self.node_parent_cache[node.uuid]].remove(node)
-            self.child_id_cache[self.node_parent_cache[node.uuid]].remove(node.uuid)
+            try:
+                self.child_cache[self.node_parent_cache[node.uuid]].remove(node)
+            except ValueError:
+                pass
+            except KeyError:
+                pass
+
+            # We have a sitation where we seem to fail here?
+            try:
+                self.child_id_cache[self.node_parent_cache[node.uuid]].remove(node.uuid)
+            except ValueError:
+                pass
+            except KeyError:
+                pass
+            del self.node_parent_cache[node.uuid]
 
         self.content_parts_cache.pop(node.uuid, None)
         self.feature_cache.pop(node.uuid, None)
@@ -597,14 +807,23 @@ class PersistenceManager(object):
 
             for child_id in child_ids:
                 child_node = self.node_cache.get_obj(child_id)
-                new_children.append(child_node) if child_node is not None else new_children.append(
-                    self.get_node(child_id))
+
+                if child_node is not None:
+                    new_children.append(child_node)
+                else:
+                    new_children.append(self.get_node(child_id))
 
             self.child_cache[node.uuid] = sorted(new_children, key=lambda x: x.index)
             self.child_id_cache[node.uuid] = set(child_ids)
 
         return self.child_cache[node.uuid]
 
+    def update_node(self, node):
+        # We need to also update the parent
+        self.node_parent_cache[node.uuid] = node._parent_uuid
+
+        self._underlying_persistence.update_node(node)
+
     def update_content_parts(self, node, content_parts):
         self.content_parts_cache[node.uuid] = content_parts
 
@@ -623,8 +842,10 @@ class PersistenceManager(object):
     def remove_feature(self, node, feature_type, name):
 
         features = self.get_features(node)
+        self._underlying_persistence.remove_feature(node, feature_type, name)
         new_features = [i for i in features if not (i.feature_type == feature_type and i.name == name)]
         self.feature_cache[node.uuid] = new_features
+        self.node_cache.add_obj(node)
 
     def get_features(self, node):
 
@@ -640,4 +861,5 @@ class PersistenceManager(object):
             features = self._underlying_persistence.get_features(node)
             self.feature_cache[node.uuid] = features
 
+        self.node_cache.add_obj(node)
         self.feature_cache[node.uuid].append(feature)
